// Copyright (c) Microsoft. All rights reserved.
// Licensed under the MIT license. See LICENSE file in the project root for full license information.

clear;

load "Params.m";

p:=2^18*3^13-1;

/*

This file performs computations as described in Section 5 of the paper.  

Note that various checks of correctness and point normalisations (that would not be used in an actual implementation) are performed throughout.

*/


alpha,E_alpha,Fp,Fp2<i>,poly:=SupersingularMontgomeryCurve(p);
K:=SupersingularKummerSurfaceFromMontgomeryCurve(alpha);

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

normalise:=function(P)

	/*

	Normalizes Kummer points in P^3, mainly for equality and correctness checking. Sets the 4th coordinate as 1, unless there is a zero coordinate, in which case it's a 2-torsion point, where we set the 1st coordinate as 1.

	INPUT: - P: Point in P^3, represented as a 4-tuple
	OUTPUT: - Normalised point equivalent to P

	*/

	if &*P ne 0 then
		return [P[1]/P[4],P[2]/P[4],P[3]/P[4],P[4]/P[4]];
	else 
		return [P[1]/P[1],P[2]/P[1],P[3]/P[1],P[4]/P[1]];
	end if;

end function;

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

OnKummer:=function(P,K)

	/*

	Checks if a given point is on a given Kummer surface.

	INPUTS: - P: Point in P^3, represented as a 4-tuple
 			- K: Kummer surface
	OUTPUT: - boolean: true if P lies on K, false otherwise

	*/

	return 
		K[1][1]*(P[1]*P[2]*P[3]*P[4])
		eq
		(
		(P[1]^2+P[2]^2+P[3]^2+P[4]^2)
		-K[1][2]*(P[1]+P[2])*(P[3]+P[4])
		-K[1][3]*(P[1]*P[2]+P[3]*P[4])
		)^2;

end function;

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

DBL:=function(P,K);

	/*

	(Pseudo-)doubling of a Kummer point

	INPUTS: - P: Point in P^3, represented as a 4-tuple, lying on...
 			- K: Kummer surface
	OUTPUT: - [2]P 

	*/

	P:=H(P);
	P:=S(P);
	P:=C(P,K[4]);
	P:=H(P);
	P:=S(P);
	P:=C(P,K[3]);

	if OnKummer(P,K) then
		return normalise(P);
	else
		return "somethings wrong";
	end if;
	
end function;

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

eDBLs:=function(P,K,e);

	/*

	Repeated (Pseudo-)doublings of a Kummer point

	INPUTS: - P: Point in P^3, represented as a 4-tuple, lying on...
 			- K: Kummer surface
 			- e: an integer
	OUTPUT: - [2^e]P 
	
	*/

	for j:=1 to e do
		P:=DBL(P,K);
	end for;

	if OnKummer(P,K) then
		return normalise(P);
	else
		return "somethings wrong";
	end if;
	
end function;

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

DBLADD:=function(P,Q,R,K);

	/*

	(Pseudo-)addition of Kummer points

	INPUTS: - P,Q,R: Points in P^3, all represented as a 4-tuple, where R=(Q+-P) lying on...
 			- K: Kummer surface
	OUTPUT: - (Q-+P)
	
	*/

	P:=H(P); U:=H(Q);
	
	Q:=C(P,K[4]);
	P:=C(P,Q);
	Q:=C(Q,U);

	P:=H(P); Q:=H(Q);
	P:=S(P); Q:=S(Q);

	P:=C(P,K[3]); Q:=C(Q,R);

	if OnKummer(P,K) and OnKummer(Q,K) then
		return P,Q;
	else
		return "somethings wrong";
	end if;
	
end function;

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

KummerScalar:=function(P,K,n);

	/*

	Scalar multiplication of Kummer points

	INPUTS: - P: Point in P^3, represented as a 4-tuple, lying on...
 			- K: Kummer surface
 			- n: a positive integer scalar
	OUTPUT: - [n]P
	
	*/

	bits:=IntegerToSequence(n,2);
	
	Pm:=P; Pp:=DBL(P,K); 
	R:=Invert4Constants(P); 

	for i:=#bits-1 to 1 by -1 do
		if bits[i] eq 0 then
			Pm,Pp:=DBLADD(Pm,Pp,R,K);
		else
			Pp,Pm:=DBLADD(Pp,Pm,R,K);
		end if;
	end for;
	
	if OnKummer(Pm,K) then
		return normalise(Pm);
	else
		return "somethings wrong";
	end if;
	
end function;

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

RandomKummerPoint:=function(K)  

	/*

	Computes a random point on a given Kummer surface

	INPUTS: - K: Kummer surface
	OUTPUT: - a random point P on K
	
	*/

	 repeat
		 x := Random(Fp);
	  	 y := Random(Fp);
	     z := Random(Fp);
	     Poly<T>:=PolynomialRing(Fp);
	     Eq := K[1][1]*(x*y*z*T)-((x^2+y^2+z^2+T^2)-K[1][2]*(x+y)*(z+T)-K[1][3]*(x*y+z*T))^2;
	     test, t := HasRoot(Eq);
     until test;

     return [x,y,z,t];

end function;

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

RandomFullOrderPoint:=function(K);

	/*

	Computes a point on a given Kummer surface whose order is maximal. Used to illustrate various SIDH-like computations. Note that it's not completely random as we force it to lie above special kernels. See the paper.

	INPUTS: - K: Kummer surface
	OUTPUT: - a point P on K
	
	*/

	tau:=(K[1][2]+Random([-1,1])*Sqrt(K[1][2]^2-4))/2; //non-zero coordinate of various 2-torsion points on K
	
	repeat
		P:=RandomKummerPoint(K);
		Q:=KummerScalar(P,K,((p+1) div 4)); //maximum order
	until normalise(Q) in {[1,0,tau,0], [1,0,tau,0], [1,0,tau,0], [1,0,tau,0]};
	
	return P;

end function;

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

IsogenousSquaredThetasFromTorsionKernel:=function(P2,P4)

	/*

	Given a special point of order 2 and a point of order 4 lying above it, this function computes the squared theta constants of the image Kummer surface corresponding to the kernel specified by the torsion points (see the paper), as well as constants used to compute the image of any point under the associated isogeny.

	INPUTS: - P2: a Kummer point of order 2 in one of the two (2,2)-subgroups of interest
			- P4: a Kummer point of order 4 such that P2=[2]P4 on K
	OUTPUT: - mus: a 4-tuple of the squared theta coordinates of the image Kummer
			- pis: some constants used in the isogeny evaluation (next) function, i.e., in the special scaling C. See the paper.
	*/

	HP2:=H(P2); HP4:=H(P4); //save on H(P2)

	assert P2[3] ne 0;

	pi1:=HP2[2]*HP4[3];
	pi2:=HP4[3]*HP2[1];
	pi3:=HP2[2]*HP4[1]; 
	pi4:=pi3;

	/*
	This has been commented out because our kernels are always of the form [1,0,tau,0], rather than [1,0,0,tau]. Other setups and instantiations may find the kernel elements to be of the form [1,0,0,tau], and instead want the "else" part of the code below. See the paper for more details. 
	*/

	/*
	if P2[3] ne 0 then
		pi1:=HP2[2]*HP4[3];
		pi2:=HP4[3]*HP2[1];
		pi3:=HP2[2]*HP4[1]; 
		pi4:=pi3;
	else
		pi1:=HP2[2]*HP4[4];
		pi2:=HP4[4]*HP2[1];
		pi3:=HP2[2]*HP4[1]; 
		pi4:=pi3;
	end if;
	*/

	pis:=[pi1,pi2,pi3,pi4];
	mus:=C(HP2,pis);
	mus:=H(mus);
	mus:=S(mus);

	return mus,pis;

end function;

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

IsogenyKummerPoint:=function(Q,pis)

	/*

	Evaluating the (2,2) isogeny from the above function on a general point on the domain Kummer. 

	INPUTS: - Q: any non-kernel point on a given Kummer K
	OUTPUT: - varphi(Q): the image of varphi on Qs

	*/

	Q:=H(Q);
	Q:=C(Q,pis);
	Q:=H(Q);
	Q:=S(Q);

	return Q;

end function;

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

/*

In what follows we perform a simple scalar-multiplication-based loop to compute a chain of Richelot isogenies on Kummers. We move the input point P through subsequent isogenies and illustrate its order being decreased by a factor of 2 each time. We also push a random point Q through the isogenies. 

This is merely to illustrate that the Kummer approach is compatible with state-of-the-art SIDH implementations. Of course, a real implementation would avoid various checks/normalizations, would work out an optimal binary-tree strategy for the main loop, etc.

See the paper for more details. 

*/


cofactor:=p+1;
repeat
	cofactor:=cofactor div 2;
until IsOdd(cofactor);

P:= RandomFullOrderPoint(K);
Q:=	RandomKummerPoint(K);
P:= KummerScalar(P,K,cofactor); 
N:=((p+1) div 2) div cofactor; //P has exact order N

assert #PrimeDivisors(N) eq 1 and IsEven(N);
assert normalise(KummerScalar(P,K,N)) eq normalise(K[2]);

e:=Factorization(N)[1][2]; //N=2^e

for l:=e to 2 by -1 do

	P4:=eDBLs(P,K,l-2);
	P2:=DBL(P4,K);

	mus,pis:=IsogenousSquaredThetasFromTorsionKernel(P2,P4);
	P:=IsogenyKummerPoint(P,pis);
	Q:=IsogenyKummerPoint(Q,pis);
	K:=SquaredKummerFromSquaredThetas(mus);
	assert OnKummer(P,K);
	assert OnKummer(Q,K);

end for;

///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////
///////////////////////////////////////////////////////////////

"done";
